#include <amsmf/Encoding.h>

namespace AMSMF
{

namespace FMT
{

namespace SMF
{

namespace MSG
{

int
lookup_msg_length(U8 stat)
{
	static const U32 systable[] = {
		/* 0xF0 */ 0, /* SYSMSG_EXCL */
		/* 0xF1 */ SysMTCQuaterFrame::LENGTH,
		/* 0xF2 */ SysSongPos::LENGTH,
		/* 0xF3 */ SysSongSel::LENGTH,
		/* 0xF4 */ 0,
		/* 0xF5 */ 0,
		/* 0xF6 */ SysTuneReq::LENGTH,
		/* 0xF7 */ 0, /* SYSMSG_EOEX */
		/* 0xF8 */ SysTimingClock::LENGTH,
		/* 0xF9 */ 0,
		/* 0xFA */ SysStart::LENGTH,
		/* 0xFB */ SysContinue::LENGTH,
		/* 0xFC */ SysStop::LENGTH,
		/* 0xFD */ 0,
		/* 0xFE */ SysActiveSensing::LENGTH,
		/* 0xFF */ SysReset::LENGTH,
	};
	static const U32 chtable[] = {
		/* 0x8x */ ChNoteOff::LENGTH,
		/* 0x9x */ ChNoteOn::LENGTH,
		/* 0xAx */ ChPolyKeyPressure::LENGTH,
		/* 0xBx */ ChControlChange::LENGTH,
		/* 0xCx */ ChProgramChange::LENGTH,
		/* 0xDx */ ChChannelPressure::LENGTH,
		/* 0xEx */ ChPitchBend::LENGTH,
	};
	if (!is_status_byte(stat))
		return 0;
	if (is_system_msg(stat))
		return systable[(int)(stat & 0x0F)];
	return chtable[(int)(stat >> 4) - 8];
}

const char *
key_sign_name_(S8 ns)
{
	static const char *names[] = {
		/* -7 */ "Cb",
		/* -6 */ "Gb",
		/* -5 */ "Db",
		/* -4 */ "Ab",
		/* -3 */ "Eb",
		/* -2 */ "Bb",
		/* -1 */ "F",
		/*  0 */ "C",
		/*  1 */ "G",
		/*  2 */ "D",
		/*  3 */ "A",
		/*  4 */ "E",
		/*  5 */ "B",
		/*  6 */ "F#",
		/*  7 */ "C#",
	};
	if (ns < -7 || ns > 7)
		return NULL;
	return names[(int)(ns+7)];
}

const char *
major_minor_name_(U8 val)
{
	switch (val) {
	case 0:
		return "major";
	case 1:
		return "minor";
	}
	return NULL;
};

} // MSG

} // SMF

} // FMT

} // AMSMF
