#ifndef AMT_SCOPED_LOCKER_H_
#define AMT_SCOPED_LOCKER_H_

#include <amt/Locker.h>

namespace AMT
{

namespace Locker
{

template <typename T_, class U_ = SimpleLocker<T_> >
class Scoped
{
	AMT_NOCOPY(Scoped);
private:
	struct Impl_ : public U_
	{
		T_ *m_ptr;
		Impl_(T_ *obj, bool locked) : m_ptr(obj)
		{
			if (obj && !locked) U_::lock(*obj);
		}
		Impl_() : m_ptr(NULL) { }
		void unlock()
		{
			if (m_ptr) {
				U_::unlock(*m_ptr);
				m_ptr = NULL;
			}
		}
		void renounce() { m_ptr = NULL; }
		void reset(T_ *obj, bool locked)
		{
			this->unlock();
			m_ptr = obj;
			if (obj && !locked) U_::lock(*obj);
		}
		~Impl_() { this->unlock(); }
	};
	Impl_ m_impl;
public:
	explicit Scoped(T_ &obj, bool locked=false) : m_impl(&obj, locked) { }
	explicit Scoped(T_ *obj, bool locked=false) : m_impl(obj, locked) { }
	Scoped() { }
	~Scoped() { }
	void unlock() { m_impl.unlock(); }
	void renounce() { m_impl.renounce(); }
	void reset(T_ &obj, bool locked=false) { m_impl.reset(&obj, locked); }
	void reset(T_ *obj, bool locked=false) { m_impl.reset(obj, locked); }
};

}

}

#endif
