//
//
//

#ifndef AMT_SORTED_DOUBLE_LINK_H_
#define AMT_SORTED_DOUBLE_LINK_H_

#include <amt/Defines.h>
#include <amt/Assert.h>

#include <amt/IntrusiveDoubleLink.h>

namespace AMT
{

namespace SortedDoubleLink
{

template <class, class> class Node;
template <class, class> class Container;

template <class T_, class A_>
class WrappedAccess_ : public A_
{
	AMT_BEGIN_CONTRACT(main)
	{
		Node<T_, A_> *n AMT_UNUSED = A_::get_node((T_ *)NULL);
		A_::dispose((Container<T_, A_> *)NULL);
		int i AMT_UNUSED =
		    A_::compare(*(const T_ *)NULL, *(const T_ *)NULL);
	}
	typedef WrappedAccess_ WA_;
	typedef IntrusiveDoubleLink::Container<T_, WA_> BaseContainer_;
protected:
	static void dispose(BaseContainer_ *) { AMT_CONFIRM_CONTRACT(main); }
	static void dispose(Container<T_, A_> *c) { A_::dispose(c); }
};

template <class T_>
struct SimpleAccessor
{
	typedef SimpleAccessor<T_> A_;
	typedef Node<T_, A_> N_;
	typedef typename N_::Container C_;
	static N_ *get_node(T_ *p) { return p; }
	static const N_ *get_node(const T_ *p) { return p; }
	static void dispose(C_ *) { }
	template <typename K1_, typename K2_>
	static int compare(const K1_ &k1, const K2_ &k2)
	{
		return T_::compare(k1, k2);
	}
};


template <class T_, class A_ = SimpleAccessor<T_> >
class Container
	: private IntrusiveDoubleLink::Container<T_, WrappedAccess_<T_, A_> >
{
	AMT_NOCOPY(Container);
public:
	typedef SortedDoubleLink::Node<T_, A_> Node;
private:
	// accessor for Base
	typedef WrappedAccess_<T_, A_> WA_;
	typedef IntrusiveDoubleLink::Container<T_, WA_ > Super_;
protected:
	class Forward_ : public Super_::Forward_
	{
		friend class Container;
	protected:
		using Super_::Forward_::compare;
	};
	class Reverse_ : public Super_::Reverse_
	{
		friend class Container;
	protected:
		template <typename Ord1_, typename Ord2_>
		static int compare(const Ord1_ &o1, const Ord2_ &o2)
		{
			return -Super_::Reverse_::compare(o1, o2);
		}
	};
	template <class Ac_>
	void confirm_assumed_after_(const char *caller, T_ *n, T_ *m) const
	{
#ifdef AMT_NEED_ASSERTION_FOR_LIBAMT_USAGE
		T_ *next;
		if (n) {
			AMT_ASSERT_LIBAMT_U_(
				caller, Ac_::compare(*n, *m) <= 0);
			next = Ac_::get_next_(n);
		} else
			next = Ac_::get_head_(this);
		if (next)
			AMT_ASSERT_LIBAMT_U_(
				caller, Ac_::compare(*m, *next) <= 0);
#endif
	}
	void confirm_after_(const char *caller, T_ *n, T_ *m) const
	{ confirm_assumed_after_<Forward_>(caller, n, m); }
	void confirm_before_(const char *caller, T_ *n, T_ *m) const
	{ confirm_assumed_after_<Reverse_>(caller, n, m); }
private:
	template <class Ac_, typename Ord_, typename Pred_>
	T_ *find_assumed_after_bw__(
		const Ord_ &ord, T_ *n, const Pred_ &p) const
	{
		for (;;) {
			T_ *prv =
			    this->Super_::template assumed_prev_pred_<
			    Ac_, Pred_>(n, p);
			if (!prv || Ac_::compare(*prv, ord) <= 0)
				break;
			n = prv;
		}
		return n;
	}
	template <class Ac_, typename Ord_, typename Pred_>
	T_ *find_assumed_after_fw__(
		const Ord_ &ord, T_ *n, const Pred_ &p) const
	{
		while ((n = this->Super_::template assumed_next_pred_<
			Ac_, Pred_>(n, p)) && Ac_::compare(*n, ord) <= 0)
			;
		return n;
	}
	template <class Ac_, typename Ord_, typename Pred_>
	T_ *find_assumed_after_(const Ord_ &ord, T_ *n, const Pred_ &p) const
	{
		n = n ? n :
		    this->Super_::template assumed_head_pred_<Ac_, Pred_>(p);
		if (n) {
			if (Ac_::compare(*n, ord) > 0) {
				n = this->find_assumed_after_bw__<
				    Ac_, Ord_, Pred_>(ord, n, p);
			} else {
				n = this->find_assumed_after_fw__<
				    Ac_, Ord_, Pred_>(ord, n, p);
			}
		}
		return n;
	}
	template <class Ac_, typename Ord_, typename Pred_>
	T_ *find_assumed_last_(const Ord_ &ord, T_ *n, const Pred_ &p) const
	{
		T_ *nxt = this->find_assumed_after_<Ac_, Ord_, Pred_>(ord,
								      n, p);
		n = this->template assumed_prev_pred_<Ac_, Pred_>(nxt, p);
		if (!n || Ac_::compare(*n, ord) != 0)
			n = nxt;

		return n;
	}
	template <class Ac_>
	void swap_assumed_next_(T_ *base)
	{
		AMT_ASSERT_LIBAMT_U_("swap_assumed_next_",
				       !Ac_::get_next_(base) ||
				       Ac_::compare(*Ac_::get_next_(base),
						    *base) == 0);
		Super_::template swap_assumed_next_<Ac_>(base);
	}
public:
	~Container() { }
	Container() { }
	using Super_::head;
	using Super_::tail;
	using Super_::next;
	using Super_::prev;
	using Super_::remove;
	using Super_::remove_head;
	using Super_::remove_tail;
	using Super_::each;
	using Super_::each_reverse;
	using Super_::remove_all;
	T_ *insert_after(T_ *base, T_ *n)
	{
		confirm_after_("insert_after", base, n);
		return Super_::insert_after(base, n);
	}
	T_ *insert_before(T_ *base, T_ *n)
	{
		confirm_before_("insert_before", base, n);
		return Super_::insert_before(base, n);
	}
	T_ *insert_head(T_ *n)
	{
		confirm_before_("insert_head", this->head(), n);
		return Super_::insert_head(n);
	}
	T_ *insert_tail(T_ *n)
	{
		confirm_after_("insert_tail", this->tail(), n);
		return Super_::insert_tail(n);
	}
	template <typename Ord_>
	T_ *find_after(const Ord_ &ord, T_ *n=NULL) const
	{ return find_assumed_after_<Forward_, Ord_, Predicates::AlwaysTrue>(
		ord, n, Predicates::AlwaysTrue()); }
	template <typename Ord_>
	T_ *find_before(const Ord_ &ord, T_ *n=NULL) const
	{ return find_assumed_after_<Reverse_, Ord_, Predicates::AlwaysTrue>(
		ord, n, Predicates::AlwaysTrue()); }
	template <typename Ord_>
	T_ *find_last(const Ord_ &ord, T_ *n=NULL) const
	{ return find_assumed_last_<Forward_, Ord_, Predicates::AlwaysTrue>(
		ord, n, Predicates::AlwaysTrue()); }
	template <typename Ord_>
	T_ *find_first(const Ord_ &ord, T_ *n=NULL) const
	{ return find_assumed_last_<Reverse_, Ord_, Predicates::AlwaysTrue>(
		ord, n, Predicates::AlwaysTrue()); }
	template <typename Ord_, typename Pred_>
	T_ *find_after(const Ord_ &ord, const Pred_ &p, T_ *n=NULL) const
	{ return find_assumed_after_<Forward_, Ord_, Pred_>(ord, n, p); }
	template <typename Ord_, typename Pred_>
	T_ *find_before(const Ord_ &ord, const Pred_ &p, T_ *n=NULL) const
	{ return find_assumed_after_<Reverse_, Ord_, Pred_>(ord, n, p); }
	template <typename Ord_, typename Pred_>
	T_ *find_last(const Ord_ &ord, const Pred_ &p, T_ *n=NULL) const
	{ return find_assumed_last_<Forward_, Ord_, Pred_>(ord, n, p); }
	template <typename Ord_, typename Pred_>
	T_ *find_first(const Ord_ &ord, const Pred_ &p, T_ *n=NULL) const
	{ return find_assumed_last_<Reverse_, Ord_, Pred_>(ord, n, p); }
	T_ *insert_suitable_first(T_ *n, T_ *base=NULL)
	{ return insert_after(find_before(*n, base), n); }
	T_ *insert_suitable_last(T_ *n, T_ *base=NULL)
	{ return insert_before(find_after(*n, base), n); }
	void swap_next(T_ *base) { swap_assumed_next_<Forward_>(base); }
	void swap_prev(T_ *base) { swap_assumed_next_<Reverse_>(base); }
	void swap(T_ *n1, T_ *n2)
	{
		AMT_ASSERT_LIBAMT_U_("swap", Forward_::compare(*n1, *n2) == 0);
		Super_::swap(n1, n2);
	}
};

template <class T_, class A_ = SimpleAccessor<T_> >
class Node : public IntrusiveDoubleLink::Node<T_, WrappedAccess_<T_, A_> >
{
public:
	typedef A_ Accessor;
	typedef SortedDoubleLink::Container<T_, A_> Container;
};

}

}

#endif
